@@ -1,7 +1,5 @@
 import json
-import logging
 import threading
-import time
 from concurrent.futures import ThreadPoolExecutor, wait
 from typing import Optional
 
@@ -11,11 +9,12 @@ from .statsig_errors import StatsigValueError, StatsigNameError
 from .statsig_network import _StatsigNetwork
 from .statsig_options import StatsigOptions
 from .thread_util import spawn_background_thread, THREAD_JOIN_TIMEOUT
+from .utils import logger
 
 RULESETS_SYNC_INTERVAL = 10
 IDLISTS_SYNC_INTERVAL = 60
 STORAGE_ADAPTER_KEY = "statsig.cache"
-
+SYNC_OUTDATED_MAX_S = 120
 
 def _is_specs_json_valid(specs_json):
     if specs_json is None or specs_json.get("time") is None:
@@ -35,7 +34,7 @@ class _SpecStore:
         self.last_update_time = 0
         self.initial_update_time = 0
         self.init_reason = EvaluationReason.uninitialized
-
+        self._initialized = False
         self._network = network
         self._options = options
         self._statsig_metadata = statsig_metadata
@@ -44,21 +43,25 @@ class _SpecStore:
         self._executor = ThreadPoolExecutor(options.idlist_threadpool_size)
         self._background_download_configs = None
         self._background_download_id_lists = None
+        self._sync_failure_count = 0
 
-        self._configs = dict()
-        self._gates = dict()
-        self._layers = dict()
-        self._experiment_to_layer = dict()
+        self._configs = {}
+        self._gates = {}
+        self._layers = {}
+        self._experiment_to_layer = {}
 
-        self._id_lists = dict()
+        self._id_lists = {}
 
         if not options.local_mode:
             self._initialize_specs()
             self.initial_update_time = -1 if self.last_update_time == 0 else self.last_update_time
-            self._download_id_lists()
 
             self.spawn_bg_threads_if_needed()
 
+            self._download_id_lists()
+
+        self._initialized = True
+
     def is_ready_for_checks(self):
         return self.last_update_time != 0
 
@@ -76,6 +79,7 @@ class _SpecStore:
         if self._options.local_mode:
             return
 
+        self._executor.shutdown(wait=False)
         self._background_download_configs.join(THREAD_JOIN_TIMEOUT)
         self._background_download_id_lists.join(THREAD_JOIN_TIMEOUT)
 
@@ -109,11 +113,12 @@ class _SpecStore:
     def _initialize_specs(self):
         if self._options.data_store is not None:
             if self._options.bootstrap_values is not None:
-                logging.getLogger('statsig.sdk').warning(
+                logger.warning(
                     "data_store gets priority over bootstrap_values. bootstrap_values will be ignored")
 
             self._load_config_specs_from_storage_adapter()
             if self.last_update_time == 0:
+                self._log_process("Retrying with network...")
                 self._download_config_specs()
 
         elif self._options.bootstrap_values is not None:
@@ -123,11 +128,13 @@ class _SpecStore:
             self._download_config_specs()
 
     def _process_specs(self, specs_json) -> bool:
+        self._log_process("Processing specs...")
         if not _is_specs_json_valid(specs_json):
+            self._log_process("Failed to process specs")
             return False
 
         def get_parsed_specs(key: str):
-            parsed = dict()
+            parsed = {}
             for spec in specs_json.get(key, []):
                 spec_name = spec.get("name")
                 if spec_name is not None:
@@ -138,7 +145,7 @@ class _SpecStore:
         new_configs = get_parsed_specs("dynamic_configs")
         new_layers = get_parsed_specs("layer_configs")
 
-        new_experiment_to_layer = dict()
+        new_experiment_to_layer = {}
         layers_dict = specs_json.get("layers", {})
         for layer_name in layers_dict:
             experiments = layers_dict[layer_name]
@@ -154,6 +161,7 @@ class _SpecStore:
         if callable(self._options.rules_updated_callback):
             self._options.rules_updated_callback(json.dumps(specs_json))
 
+        self._log_process("Done processing specs")
         return True
 
     def _bootstrap_config_specs(self):
@@ -172,24 +180,37 @@ class _SpecStore:
                 self.init_reason = EvaluationReason.bootstrap
         except ValueError:
             # JSON decoding failed, just let background thread update rulesets
-            logging.getLogger('statsig.sdk').exception(
+            logger.exception(
                 'Failed to parse bootstrap_values')
             return
 
     def _spawn_bg_download_config_specs(self):
-        self._background_download_configs = spawn_background_thread(self._sync, (
-            self._download_config_specs, self._options.rulesets_sync_interval or RULESETS_SYNC_INTERVAL
-        ), self._error_boundary)
+        self._background_download_configs = spawn_background_thread(
+            self._sync,
+            (self._download_config_specs, self._options.
+             rulesets_sync_interval or RULESETS_SYNC_INTERVAL),
+            self._error_boundary)
 
     def _download_config_specs(self):
+        self._log_process("Loading specs from network...")
+        log_on_exception = not self._initialized
+        if self._sync_failure_count * self._options.rulesets_sync_interval > 120:
+            log_on_exception = True
+            self._sync_failure_count = 0
+
         specs = self._network.post_request("download_config_specs", {
             "statsigMetadata": self._statsig_metadata,
             "sinceTime": self.last_update_time,
-        })
+        }, log_on_exception)
+
+        if specs is None:
+            self._sync_failure_count += 1
+            return
 
         if not _is_specs_json_valid(specs):
             return
 
+        self._log_process("Done loading specs")
         if self._process_specs(specs):
             self._save_to_storage_adapter(specs)
             self.init_reason = EvaluationReason.network
@@ -207,6 +228,7 @@ class _SpecStore:
         self._options.data_store.set(STORAGE_ADAPTER_KEY, json.dumps(specs))
 
     def _load_config_specs_from_storage_adapter(self):
+        self._log_process("Loading specs from adapter")
         if self._options.data_store is None:
             return
 
@@ -216,13 +238,16 @@ class _SpecStore:
 
         cache = json.loads(cache_string)
         if not isinstance(cache, dict):
-            logging.getLogger('statsig.sdk').warning("Invalid type returned from StatsigOptions.data_store")
+            logger.warning(
+                "Invalid type returned from StatsigOptions.data_store")
             return
 
         adapter_time = cache.get("time", None)
-        if not isinstance(adapter_time, int) or adapter_time < self.last_update_time:
+        if not isinstance(adapter_time,
+                          int) or adapter_time < self.last_update_time:
             return
 
+        self._log_process("Done loading specs")
         if self._process_specs(cache):
             self.init_reason = EvaluationReason.data_adapter
 
@@ -243,10 +268,10 @@ class _SpecStore:
             workers = []
 
             for list_name in server_id_lists:
-                server_list = server_id_lists.get(list_name, dict())
+                server_list = server_id_lists.get(list_name, {})
                 url = server_list.get("url", None)
                 size = server_list.get("size", 0)
-                local_list = local_id_lists.get(list_name, dict())
+                local_list = local_id_lists.get(list_name, {})
 
                 new_creation_time = server_list.get("creationTime", 0)
                 old_creation_time = local_list.get("creationTime", 0)
@@ -272,6 +297,10 @@ class _SpecStore:
                 if size <= read_bytes or url == "":
                     continue
 
+                # if the main id sync is killed, we should too
+                if not self._background_download_id_lists or not self._background_download_id_lists.is_alive():
+                    return
+
                 future = self._executor.submit(
                     self._download_single_id_list,
                     url, list_name, local_list, local_id_lists, read_bytes
@@ -291,9 +320,10 @@ class _SpecStore:
         except Exception as e:
             self._error_boundary.log_exception(e)
 
-    def _download_single_id_list(self, url, list_name, local_list, all_lists, start_index):
+    def _download_single_id_list(
+            self, url, list_name, local_list, all_lists, start_index):
         resp = self._network.get_request(
-            url, headers={"Range": "bytes=%s-" % start_index})
+            url, headers={"Range": f"bytes={start_index}-"})
         if resp is None:
             return
         try:
@@ -305,7 +335,7 @@ class _SpecStore:
             if content is None:
                 return
             first_char = content[0]
-            if first_char != "+" and first_char != "-":
+            if first_char not in ('+', '-'):
                 raise StatsigNameError("Seek range invalid.")
             lines = content.splitlines()
             for line in lines:
@@ -330,3 +360,8 @@ class _SpecStore:
                 sync_func()
             except Exception as e:
                 self._error_boundary.log_exception(e)
+
+    def _log_process(self, msg, process=None):
+        if process is None:
+            process = "Initialize" if not self._initialized else "Sync"
+        logger.log_process(process, msg)
